//
// TM & (c) 2017 Lucasfilm Entertainment Company Ltd. and Lucasfilm Ltd.
// All rights reserved.  See LICENSE.txt for license.
//

#include <MaterialXRender/ImageHandler.h>

#include <MaterialXGenShader/Shader.h>
#include <MaterialXGenShader/Util.h>

#include <iostream>

namespace MaterialX
{

const string IMAGE_PROPERTY_SEPARATOR("_");
const string UADDRESS_MODE_SUFFIX(IMAGE_PROPERTY_SEPARATOR + "uaddressmode");
const string VADDRESS_MODE_SUFFIX(IMAGE_PROPERTY_SEPARATOR + "vaddressmode");
const string FILTER_TYPE_SUFFIX(IMAGE_PROPERTY_SEPARATOR + "filtertype");
const string DEFAULT_COLOR_SUFFIX(IMAGE_PROPERTY_SEPARATOR + "default");

const string ImageLoader::BMP_EXTENSION = "bmp";
const string ImageLoader::EXR_EXTENSION = "exr";
const string ImageLoader::GIF_EXTENSION = "gif";
const string ImageLoader::HDR_EXTENSION = "hdr";
const string ImageLoader::JPG_EXTENSION = "jpg";
const string ImageLoader::JPEG_EXTENSION = "jpeg";
const string ImageLoader::PIC_EXTENSION = "pic";
const string ImageLoader::PNG_EXTENSION = "png";
const string ImageLoader::PSD_EXTENSION = "psd";
const string ImageLoader::TGA_EXTENSION = "tga";
const string ImageLoader::TIF_EXTENSION = "tif";
const string ImageLoader::TIFF_EXTENSION = "tiff";
const string ImageLoader::TX_EXTENSION = "tx";
const string ImageLoader::TXT_EXTENSION = "txt";
const string ImageLoader::TXR_EXTENSION = "txr";

//
// ImageHandler methods
//

ImageHandler::ImageHandler(ImageLoaderPtr imageLoader)
{
    addLoader(imageLoader);
    _zeroImage = createUniformImage(2, 2, 4, Image::BaseType::UINT8, Color4(0.0f));

    // Generated shaders interpret 1x1 textures as invalid images.
    _invalidImage = createUniformImage(1, 1, 4, Image::BaseType::UINT8, Color4(0.0f));
}

void ImageHandler::addLoader(ImageLoaderPtr loader)
{
    if (loader)
    {
        const StringSet& extensions = loader->supportedExtensions();
        for (const auto& extension : extensions)
        {
            _imageLoaders[extension].push_back(loader);
        }
    }
}

StringSet ImageHandler::supportedExtensions()
{
    StringSet extensions;
    for (const auto& pair : _imageLoaders)
    {
        extensions.insert(pair.first);
    }
    return extensions;
}

bool ImageHandler::saveImage(const FilePath& filePath,
                             ConstImagePtr image,
                             bool verticalFlip)
{
    if (!image)
    {
        return false;
    }

    FilePath foundFilePath =  _searchPath.find(filePath);
    if (foundFilePath.isEmpty())
    {
        return false;
    }

    string extension = foundFilePath.getExtension();
    for (ImageLoaderPtr loader : _imageLoaders[extension])
    {
        bool saved = false;
        try
        {
            saved = loader->saveImage(foundFilePath, image, verticalFlip);
        }
        catch (std::exception& e)
        {
            std::cerr << "Exception in image I/O library: " << e.what() << std::endl;
        }
        if (saved)
        {
            return true;
        }
    }
    return false;
}

ImagePtr ImageHandler::acquireImage(const FilePath& filePath, bool)
{
    FilePath foundFilePath = _searchPath.find(filePath);
    string extension = stringToLower(foundFilePath.getExtension());
    bool extensionSupported = false;
    for (ImageLoaderPtr loader : _imageLoaders[extension])
    {
        extensionSupported = true;
        ImagePtr image;
        try
        {
            image = loader->loadImage(foundFilePath);
        }
        catch (std::exception& e)
        {
            std::cerr << "Exception in image I/O library: " << e.what() << std::endl;
        }
        if (image)
        {
            // Generated shaders interpret 1x1 textures as invalid images, so valid 1x1
            // images must be resized.
            if (image->getWidth() == 1 && image->getHeight() == 1)
            {
                image = createUniformImage(2, 2, image->getChannelCount(),
                                            image->getBaseType(), image->getTexelColor(0, 0));
            }

            cacheImage(foundFilePath, image);
            return image;
        }
    }

    if (!foundFilePath.isEmpty())
    {
        if (!foundFilePath.exists())
        {
            std::cerr << string("Image file not found: ") + foundFilePath.asString() << std::endl;
        }
        else if (!extensionSupported)
        {
            std::cerr << string("Unsupported image extension: ") + foundFilePath.asString() << std::endl;
        }
        else
        {
            std::cerr << string("Image loader failed to parse image: ") + foundFilePath.asString() << std::endl;
        }
    }

    cacheImage(foundFilePath, _invalidImage);
    return _invalidImage;
}

bool ImageHandler::bindImage(ImagePtr, const ImageSamplingProperties&)
{
    return false;
}

bool ImageHandler::unbindImage(ImagePtr)
{
    return false;
}

void ImageHandler::unbindImages()
{
    for (auto iter : _imageCache)
    {
        unbindImage(iter.second);
    }
}

bool ImageHandler::createRenderResources(ImagePtr, bool)
{
    return false;
}

void ImageHandler::releaseRenderResources(ImagePtr)
{
}

void ImageHandler::cacheImage(const string& filePath, ImagePtr image)
{
    _imageCache[filePath] = image;
}

ImagePtr ImageHandler::getCachedImage(const FilePath& filePath)
{
    if (_imageCache.count(filePath))
    {
        return _imageCache[filePath];
    }
    if (!filePath.isAbsolute())
    {
        for (const FilePath& path : _searchPath)
        {
            FilePath combined = path / filePath;
            if (_imageCache.count(combined))
            {
                return _imageCache[combined];
            }
        }
    }
    return nullptr;
}

void ImageHandler::clearImageCache()
{
    for (auto iter : _imageCache)
    {
        releaseRenderResources(iter.second);
    }
    _imageCache.clear();
}

//
// ImageSamplingProperties methods
//

void ImageSamplingProperties::setProperties(const string& fileNameUniform,
                                            const VariableBlock& uniformBlock)
{
    const int INVALID_MAPPED_INT_VALUE = -1; // Any value < 0 is not considered to be invalid

    // Get the additional texture parameters based on image uniform name
    // excluding the trailing "_file" postfix string
    string root = fileNameUniform;
    size_t pos = root.find_last_of(IMAGE_PROPERTY_SEPARATOR);
    if (pos != string::npos)
    {
        root = root.substr(0, pos);
    }

    const string uaddressmodeStr = root + UADDRESS_MODE_SUFFIX;
    const ShaderPort* port = uniformBlock.find(uaddressmodeStr);
    ValuePtr intValue = port ? port->getValue() : nullptr;
    uaddressMode = ImageSamplingProperties::AddressMode(intValue && intValue->isA<int>() ? intValue->asA<int>() : INVALID_MAPPED_INT_VALUE);

    const string vaddressmodeStr = root + VADDRESS_MODE_SUFFIX;
    port = uniformBlock.find(vaddressmodeStr);
    intValue = port ? port->getValue() : nullptr;
    vaddressMode = ImageSamplingProperties::AddressMode(intValue && intValue->isA<int>() ? intValue->asA<int>() : INVALID_MAPPED_INT_VALUE);

    const string filtertypeStr = root + FILTER_TYPE_SUFFIX;
    port = uniformBlock.find(filtertypeStr);
    intValue = port ? port->getValue() : nullptr;
    filterType = ImageSamplingProperties::FilterType(intValue && intValue->isA<int>() ? intValue->asA<int>() : INVALID_MAPPED_INT_VALUE);

    const string defaultColorStr = root + DEFAULT_COLOR_SUFFIX;
    port = uniformBlock.find(defaultColorStr);
    ValuePtr colorValue = port ? port->getValue() : nullptr;
    if (colorValue)
    {
        mapValueToColor(colorValue, defaultColor);
    }
}

} // namespace MaterialX
